from abc import ABC, abstractmethod
from typing import List, TypedDict, Optional

class QuestionAnswererResponse(TypedDict):
    answer: str
    explanation: Optional[str]

class QuestionAnswerer(ABC):
    
        @abstractmethod
        def answer(self, questions: List[str], context: str) -> QuestionAnswererResponse:
            pass


from .metric import Metric


class ContradictionScore(Metric):
    """
    Metric to evaluate the degree of contradiction between the answers obtained from
    a summary and the original document. It captures the percentage of questions that
    received contradictory answers between the summary and the document, with neither
    being 'Unknown'. A high score suggests the summary might be contradicting the
    original document's content.

    Attributes:
        answers_src (dict): Answers derived from the original document.
        answers_sum (dict): Answers derived from the summary.
        n_questions (int): Number of questions posed.
    """

    @staticmethod
    def _compute_metric(answers_src, answers_sum, questions):
        """
        Compute the number of contradictions between answers derived from the document
        and the summary.

        Args:
            answers_src (dict): Answers based on the original document.
            answers_sum (dict): Answers based on the summary.

        Returns:
            int: Number of contradictions.
        """
        answers_src_ls = list(answers_src.values())
        answers_sum_ls = list(answers_sum.values())

        n_contradiction = 0
        cont_questions = []

        for idx, (ans_src, ans_sum) in enumerate(zip(answers_src_ls, answers_sum_ls)):
            if (
                ans_src.strip().lower() in ["yes", "no"]
                and ans_src.strip().lower() != ans_sum.strip().lower()
            ):
                n_contradiction += 1
                cont_question = questions[idx]
                cont_questions.append(f"{cont_question}")

        return n_contradiction, cont_questions

    @staticmethod
    def compute(answers_src, answers_sum, questions, n_questions):
        """
        Compute the contradiction score by normalizing the number of contradictions by
        the total number of questions.

        Args:
            answers_src (dict): Answers based on the original document.
            answers_sum (dict): Answers based on the summary.
            n_questions (int): Total number of questions.

        Returns:
            float: Contradiction score.
        """
        n_contradiction, cont_questions = ContradictionScore._compute_metric(
            answers_src, answers_sum, questions
        )
        explanation = cont_questions
        cont_score = n_contradiction / n_questions
        return (cont_score, explanation)



import json
import jsonschema
from jsonpath_ng import parse
from jsonschema import validate
from typing import Any, Optional

class JsonHelper:
    @staticmethod
    def _extract_json(data_string: str) -> str:
        """
        Extracts a JSON string from a larger string.
        Assumes the JSON content starts with '{' and continues to the end of the input string.
        """
        try:
            start_index = data_string.index("{")
            end_index = data_string.rfind("}")
            json_string = data_string[start_index : end_index + 1]
        except Exception as e:
            json_string = data_string
        return json_string

    @staticmethod
    def _load_json_from_text(text):
        """
        Extracts and loads a JSON string from a given text.
        """
        try:
            data = json.loads(text)
        except json.decoder.JSONDecodeError:
            raise ValueError("Failed to load JSON from text")
        return data

    @staticmethod
    def extract_json_from_text(text):
        # In case you cannot handle an error, return None
        if text is None:
            return None
        response_json_format = JsonHelper._extract_json(text)
        response_json = JsonHelper._load_json_from_text(response_json_format)
        return response_json

def validate_json(json_data, schema):
    try:
        validate(instance=json_data, schema=schema)
        return True, None
    except jsonschema.exceptions.ValidationError as err:
        return False, str(err)

def extract_json_path(json_data, json_path):
    try:
        jsonpath_expr = parse(json_path)
        match = jsonpath_expr.find(json_data)
        return [match.value for match in match] if match else None
    except Exception as e:
        return None

# New and improved JsonExtractor
# - can extract top-level arrays as well
# - uses stack based approach
class JsonExtractor:
    @staticmethod
    def extract_first_json_entity(text: str) -> Optional[Any]:
        """
        Extracts the first top-level JSON entity from a given text string.

        Args:
            text (str): The input text containing JSON entities.

        Returns:
            dict or list: The first JSON object or array extracted from the text, or None if no valid JSON is found.
        """
        i = 0
        length = len(text)

        while i < length:
            if text[i] in "{[":
                start_idx = i
                stack = [text[i]]
                i += 1

                while i < length and stack:
                    if text[i] in "{[":
                        stack.append(text[i])
                    elif text[i] in "}]":
                        stack.pop()
                    i += 1

                if not stack:
                    json_str = text[start_idx:i]
                    try:
                        return json.loads(json_str)
                    except json.JSONDecodeError:
                        continue
            else:
                i += 1

        return None



from typing import Union, Dict, Any
from athina.steps import Step
import json
from jsonpath_ng import parse

class ExtractJsonPath(Step):
    """
    Step that extracts json path from text using the JsonPath provided to the step.

    Attributes:
        input_column: The row's column to extract JsonPath from.
        json_path: The JsonPath to extract from the text.
    """

    input_column: str
    json_path: str

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Extract the JsonPath from the input data."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        input_text = input_data.get(self.input_column, None)
        
        if input_text is None:
            return None

        try:
            if isinstance(input_text, dict) or isinstance(input_text, list):
                input_json = input_text
            elif isinstance(input_text, str):
                input_json =  json.loads(input_text)
            else:
                return None
            result = parse(self.json_path).find(input_json)
            
            if not result or len(result) == 0:
                result = None
            elif len(result) == 1:
                result = result[0].value
            else:
                result = [match.value for match in result]

            return {
                "status": "success",
                "data": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "data": str(e),
            }



import inspect


def get_named_parameters(func):
    """
    Get all named parameters of a function.
    """
    parameters = inspect.signature(func).parameters
    named_parameters = [param for param in parameters if parameters[param].default != inspect.Parameter.empty]
    return named_parameters

def get_named_non_default_parameters(func):
    """
    Get all named parameters without default values of a function.
    """
    parameters = inspect.signature(func).parameters
    named_non_default_parameters = [param for param in parameters if parameters[param].default == inspect.Parameter.empty]
    return named_non_default_parameters


from typing import Union, Dict, Any, Optional
from athina.steps import Step
import subprocess
import os
import json
import re
import tempfile 

class CodeExecution(Step):
    """
    Step that executes code using the code provided to the step.

    Attributes:
        code: The code to execute.
    """

    code: str
    name: Optional[str] = None

    def bandit_check(self, code: str) -> None:
        """
        Run Bandit security check on the provided code.
        """
        with tempfile.NamedTemporaryFile(delete=False, suffix=".py") as temp_file:
            temp_file.write(code.encode('utf-8'))
            temp_file_path = temp_file.name
        try:
            result = subprocess.run(
                ["bandit", "-r", temp_file_path, "-f", "json", "-c", "bandit.yml"],
                capture_output=True,
                text=True
            )
            if result.returncode != 0:
                return json.dumps(result.stdout)
        finally:
            os.remove(temp_file_path)
        return None
    
    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Extract the JsonPath from the input data."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        try:
            issues = self.bandit_check(self.code)
            if not issues:
                return { "error": "Security check failed. " + issues }
            from RestrictedPython import compile_restricted
            from RestrictedPython import safe_globals
            from RestrictedPython.Guards import safe_builtins
            from RestrictedPython.Eval import default_guarded_getitem, default_guarded_getiter
            from textatistic import Textatistic
            import editdistance
            import textdistance
            from datetime import datetime
            import time
            
            custom_builtins = safe_builtins.copy()
            custom_builtins.update({
                'type': type,
                'dict': dict,
                'list': list,
                'set': set,
                'tuple': tuple,
                'str': str,
                'int': int,
                'float': float,
                'bool': bool,
                'len': len,
                'range': range,
                'enumerate': enumerate,
                'zip': zip,
                'sorted': sorted,
                'min': min,
                'max': max,
                'sum': sum,
                'abs': abs,
                'all': all,
                'any': any,
                'isinstance': isinstance,
                'issubclass': issubclass,
                'Textatistic': Textatistic,
                'datetime': datetime,
                'Exception': Exception,
                'ValueError': ValueError,
                'TypeError': TypeError,
                'KeyError': KeyError,
                'IndexError': IndexError,
                'AttributeError': AttributeError,
                'ImportError': ImportError,
                '__import__': __import__
            })

            custom_globals = safe_globals.copy()
            custom_globals.update({
                '__builtins__': custom_builtins,
                'json': json,
                're': re,
                'editdistance': editdistance,
                'textdistance': textdistance,
                'datetime': datetime,
                'time': time,
                '_getitem_': default_guarded_getitem,
                '_getiter_': default_guarded_getiter,
                '_write_': lambda x: x
            })
            # Whitelist of allowed modules
            allowed_modules = {'json', 're', 'editdistance', 'textdistance', 'datetime', 'time'}
            def guarded_import(name, *args, **kwargs):
                if name not in allowed_modules:
                    raise ImportError(f"Importing '{name}' is not allowed")
                return __import__(name, *args, **kwargs)

            custom_builtins['__import__'] = guarded_import
            loc = {}
            byte_code = compile_restricted(self.code, '<inline>', 'exec')
            exec(byte_code, custom_globals, loc)
            result = loc['main'](**input_data)
            return {
                "status": "success",
                "data": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "data": str(e),
            }





data = [
    # Incorrect - Unfaithful
    {
        "query": "What are some successful companies that went through YC?",
        "context": ["Y Combinator has invested in companies in various fields like FinTech, Healthcare, AI, etc."],
        "response": "Airbnb, Dropbox, Stripe, Reddit, Coinbase, Instacart.",
        "expected_response": "Airbnb and Stripe are 2 of the successful companies that went through YC.",
    },
    {
        "query": "In which city is YC located?",
        "context": ["Y Combinator is located in Mountain View, California."],
        "response": "Y Combinator is located in San Francisco",
        "expected_response": "YC is located in Mountain View, California.",
    },
    # Incorrect - Insufficient Context + Unfaithful
    {
        "query": "How much equity does YC take?",
        "context": ["Y Combinator invests $500k in 200 startups twice a year."],
        "response": "YC invests $150k for 7%.",
        "expected_response": "I cannot answer this question as I do not have enough information.",
    },
    # Incorrect - Insufficient Answer
    {
        "query": "How much equity does YC take?",
        "context": ["Y Combinator invests $500k in 200 startups twice a year."],
        "response": "I cannot answer this question as I do not have enough information.",
        "expected_response": "I cannot answer this question as I do not have enough information.",
    },
    {
        "query": "Who founded YC and when was it founded?",
        "context": ["Y Combinator was founded in March 2005 by Paul Graham, Jessica Livingston, Trevor Blackwell, and Robert Tappan Morris."],
        "response": "Y Combinator was founded in 2005",
        "expected_response": "Y Combinator was founded in March 2005 by Paul Graham, Jessica Livingston, Trevor Blackwell, and Robert Tappan Morris.",
    },
    # Correct answers
    {
        "query": "Does Y Combinator invest in startups outside the US?",
        "context": ["Y Combinator invests in startups from all over the world."],
        "response": "Yes, Y Combinator invests in international startups as well as US startups.",
        "expected_response": "Yes, Y Combinator invests in startups from all over the world.",
    },
    {
        "query": "How much does YC invest in startups?",
        "context": ["YC invests $150k for 7%."],
        "response": "$150k",
        "expected_response": "YC invests $150k for 7%.",
    },
    {
        "query": "What is YC's motto?",
        "context": ["Y Combinator's motto is 'Make something people want'."],
        "response": "Make something people want",
        "expected_response": "Make something people want",
    },
]



from typing import Optional, List
from athina.evals.eval_type import FunctionEvalTypeId
from athina.evals.function.function_evaluator import FunctionEvaluator


class ContainsAny(FunctionEvaluator):
    def __init__(
        self,
        keywords: List[str],
        case_sensitive: Optional[bool] = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the ContainsAny function evaluator.

        Args:
            keywords (List[str]): List of keywords to check for in the text.
            case_sensitive (Optional[bool], optional): Whether the keyword matching should be case sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_ANY.value,
            function_arguments={"keywords": keywords, "case_sensitive": case_sensitive},
            display_name=display_name,
        )


class Regex(FunctionEvaluator):
    def __init__(
        self,
        pattern: str,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the Regex function evaluator.

        Args:
            pattern (str): The regular expression pattern to be matched in the text.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.REGEX.value,
            function_arguments={"pattern": pattern},
            display_name=display_name,
        )


class ContainsNone(FunctionEvaluator):
    def __init__(
        self,
        keywords: List[str],
        case_sensitive: bool = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the ContainsNone function evaluator.

        Args:
            keywords (str or List[str]): The keyword(s) to search for in the text.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_NONE.value,
            function_arguments={
                "keywords": keywords,
                "case_sensitive": case_sensitive,
            },
            display_name=display_name,
        )


class Contains(FunctionEvaluator):
    def __init__(
        self,
        keyword: str,
        case_sensitive: bool = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the Contains function evaluator.

        Args:
            keyword (str): The keyword to search for in the text.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS.value,
            function_arguments={
                "keyword": keyword,
                "case_sensitive": case_sensitive,
            },
            display_name=display_name,
        )


class ContainsAll(FunctionEvaluator):
    def __init__(
        self,
        keywords: List[str],
        case_sensitive: bool = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the ContainsAll function evaluator.

        Args:
            keywords (List[str]): The list of keywords to search for in the text.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_ALL.value,
            function_arguments={
                "keywords": keywords,
                "case_sensitive": case_sensitive,
            },
            display_name=display_name,
        )


class ContainsJson(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the ContainsJson function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_JSON.value,
            function_arguments={},
        )


class ContainsEmail(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the ContainsEmail function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_EMAIL.value,
            function_arguments={},
            display_name=display_name,
        )


class IsJson(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the IsJson function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.IS_JSON.value,
            function_arguments={},
            display_name=display_name,
        )


class IsEmail(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the IsEmail function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.IS_EMAIL.value,
            function_arguments={},
            display_name=display_name,
        )


class NoInvalidLinks(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the NoInvalidLinks function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.NO_INVALID_LINKS.value,
            function_arguments={},
            display_name=display_name,
        )


class ContainsLink(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the ContainsLink function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_LINK.value,
            function_arguments={},
            display_name=display_name,
        )


class ContainsValidLink(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the ContainsValidLink function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_VALID_LINK.value,
            function_arguments={},
            display_name=display_name,
        )


class Equals(FunctionEvaluator):
    def __init__(
        self,
        case_sensitive: bool = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the Equals function evaluator.

        Args:
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.EQUALS.value,
            function_arguments={
                "case_sensitive": case_sensitive,
            },
            display_name=display_name,
        )


class StartsWith(FunctionEvaluator):
    def __init__(
        self,
        substring: str,
        case_sensitive: bool = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the StartsWith function evaluator.

        Args:
            substring (str): The substring to check for at the start of the text.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.STARTS_WITH.value,
            function_arguments={
                "substring": substring,
                "case_sensitive": case_sensitive,
            },
            display_name=display_name,
        )


class EndsWith(FunctionEvaluator):
    def __init__(
        self,
        substring: str,
        case_sensitive: bool = False,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the EndsWith function evaluator.

        Args:
            substring (str): The substring to check for at the end of the text.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.ENDS_WITH.value,
            function_arguments={
                "substring": substring,
                "case_sensitive": case_sensitive,
            },
            display_name=display_name,
        )


class LengthLessThan(FunctionEvaluator):
    def __init__(self, max_length: int, display_name: Optional[str] = None):
        """
        Initialize the LengthLessThan function evaluator.

        Args:
            max_length (int): The maximum length that the text should have.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.LENGTH_LESS_THAN.value,
            function_arguments={
                "max_length": max_length,
            },
            display_name=display_name,
        )


class LengthGreaterThan(FunctionEvaluator):
    def __init__(self, min_length: int, display_name: Optional[str] = None):
        """
        Initialize the LengthGreaterThan function evaluator.

        Args:
            min_length (int): The minimum length that the text should have.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.LENGTH_GREATER_THAN.value,
            function_arguments={
                "min_length": min_length,
            },
            display_name=display_name,
        )


class ApiCall(FunctionEvaluator):
    def __init__(
        self,
        url: str,
        payload: Optional[dict] = None,
        headers: Optional[dict] = None,
        display_name: Optional[str] = None,
    ):
        """
        Initialize the ApiCall function evaluator.

        Args:
            url (str): The URL to make the API call to.
            payload (dict): The payload to be sent in the API call. response, query, context, expected_response will be added to the payload.
            headers (dict, optional): The headers to be included in the API call. Defaults to None.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.API_CALL.value,
            function_arguments={
                "url": url,
                "payload": payload,
                "headers": headers,
            },
            display_name=display_name,
        )

class LengthBetween(FunctionEvaluator):
    def __init__(self, min_length: int, max_length: int, display_name: Optional[str] = None):
        """
        Initialize the LengthBetween function evaluator.

        Args:
            min_length (int): The minimum length that the text should have.
            max_length (int): The maximum length that the text should have.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.LENGTH_BETWEEN.value,
            function_arguments={
                "min_length": min_length,
                "max_length": max_length,
            },
            display_name=display_name,
        )

class OneLine(FunctionEvaluator):
    def __init__(self, display_name: Optional[str] = None):
        """
        Initialize the OneLine function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.ONE_LINE.value,
            function_arguments={},
            display_name=display_name,
        )

class CustomCodeEval(FunctionEvaluator):
    def __init__(self, code: str, display_name: Optional[str] = None):
        """
        Initialize the Custom code evaluator.

        Args:
            code (str): The custom code to be executed.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CUSTOM_CODE_EVAL.value,
            function_arguments={
                "code": code,
            },
            display_name=display_name,
        )

class JsonSchema(FunctionEvaluator):
    def __init__(self, schema: str, display_name: Optional[str] = None):
        """
        Initialize the JsonSchema function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.JSON_SCHEMA.value,
            function_arguments={
                "schema": schema
            },
            display_name=display_name,
        )

class JsonValidation(FunctionEvaluator):
    def __init__(self, validations = None, display_name: Optional[str] = None):
        """
        Initialize the JsonValidation function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.JSON_VALIDATION.value,
            function_arguments={
                "validations": validations
            },
            display_name=display_name,
        )



import litellm
from retrying import retry
from timeout_decorator import timeout
from athina.helpers.json import JsonHelper
from athina.keys import OpenAiApiKey
from athina.interfaces.model import Model
from athina.errors.exceptions import NoOpenAiApiKeyException
from .abstract_llm_service import AbstractLlmService


class LitellmService(AbstractLlmService):
    _instance = None
    _api_key = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(LitellmService, cls).__new__(cls)
        return cls._instance

    def __init__(self, api_key):
        self._api_key = api_key

    def embeddings(self, text: str) -> list:
        """
        Fetches response from OpenAI's Embeddings API.
        """
        raise NotImplementedError

    @retry(stop_max_attempt_number=3, wait_fixed=2000)
    def chat_completion(self, messages, model, **kwargs) -> str:
        """
        Fetches response from Litellm's Completion API.
        """
        try:
            response = litellm.completion(api_key=self._api_key, model=model, messages=messages, **kwargs)
            return response.choices[0].message.content
        except Exception as e:
            print(f"Error in ChatCompletion: {e}")
            raise e

    @retry(stop_max_attempt_number=3, wait_fixed=2000)
    def chat_completion_json(self, messages, model, **kwargs) -> str:
        raise NotImplementedError

    def json_completion(self, messages, model, **kwargs):
        raise NotImplementedError
